#include <config.h>

module hmc_class
!
!= This module defines HMC algorithm
!
!   type(hmc_algorithm) :: hmc
!   call new(hmc)
!   call read(hmc)
!   call print(hmc)
!   call run(hmc)
!   call save(hmc)
!   call print_statistics(hmc)
!   call delete(hmc)
!
!== Version
!
! $Id: hmc_class.F90,v 1.19 2011/12/01 06:55:01 ishikawa Exp $
!
  use lattice_class
  use counter_class
  use error_class
  use timer_class
  use hmc_logfile_class
  use hmc_status_class
  use md_status_class
  use hmc_qcd_class
  implicit none
  private
  public :: wloops_obj
  public :: hmc_algorithm
  public :: new
  public :: delete
  public :: print
  public :: read
  public :: save
  public :: print_statistics
  public :: run

  type hmc_algorithm   !:doc-priority -100:
  !
  ! HMC program parameter
  !
    private
    type(qcd_hmc)       :: qcd
    type(md_status)     ::  md_status
    type(hmc_status)    :: hmc_status
    type(timer)         :: timing_program
    type(timer)         :: timing_hmonte
    type(timer)         :: timing_metropolis
    type(timer)         :: timing_io
    integer             :: random_number_seed = 0
    character(CHARLEN) :: file_name=""
    character(CHARLEN) :: current_directory=""
    character(CHARLEN) :: job_id=""
    character(CHARLEN) :: config_path=""
    character(CHARLEN) :: config_file_name_header=""
    character(CHARLEN) :: input_file_comment(4)
  end type

  interface new
  !
  ! Initialize/create HMC algorithm method
  !
    module procedure new_hmc
  end interface

  interface read
  !
  ! read HMC algorithm parameter/status method
  !
    module procedure read_hmc
  end interface

  interface save
  !
  ! save HMC algorithm parameter/status method
  !
    module procedure save_hmc
  end interface

  interface print
  !
  ! print out HMC algorithm parameter/status method
  !
    module procedure print_hmc
  end interface

  interface run
  !
  ! run HMC algorithm method
  !
    module procedure run_hmc
  end interface

  interface print_statistics
  !
  ! print out HMC algorithm job statistics method
  !
    module procedure print_stat_hmc
  end interface

  interface delete
  !
  ! delete HMC algorithm method
  !
    module procedure delete_hmc
  end interface

contains

subroutine new_hmc(this)
!
! Initialize and Create HMC algorithm
!
  use comlib
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  character(len=8)  :: cdate
  character(len=10) :: ctime
  integer :: iflg
  character(CHARLEN) :: str

  call tic(this%timing_program)

  call new(this%hmc_status)
  call new(this%md_status)
  call new(this%qcd,this%hmc_status)

  this%file_name=repeat(' ',len(this%file_name))
  this%current_directory=repeat(' ',len(this%current_directory))
  this%job_id=repeat(' ',len(this%job_id))

  call DATE_AND_TIME(date=cdate,time=ctime)
  iflg=0
  if (nodeid == 0) then
    if(COMMAND_ARGUMENT_COUNT()==1) then
      call GET_COMMAND_ARGUMENT(1,this%file_name)
    else
      iflg=1
    endif
  endif
#ifndef _singlePU
  call comlib_bcast(iflg,0)
#endif
  if (iflg==1) then
    write(str,'("HMC new error: ",A,I4)')__FILE__,__LINE__
    call error_stop(TRIM(str))
  endif
  if (nodeid == 0) then
    call getcwd(this%current_directory)
  endif

  write(this%job_id,'(A,1X,A)')TRIM(ADJUSTL(cdate)),TRIM(ADJUSTL(ctime))

  if (nodeid == 0) then
    write(*,'(" DATE AND TIME :",A)')TRIM(this%job_id)
  endif

#ifndef _singlePU
  call comlib_bcast(this%file_name,0)
  call comlib_bcast(this%current_directory,0)
  call comlib_bcast(this%job_id,0)
#endif

  this%config_path=repeat(' ',len(this%config_path))
  this%config_file_name_header=repeat(' ',len(this%config_file_name_header))

  return
end subroutine

subroutine delete_hmc(this)
!
! Delelete HMC algorithm
!
  use comlib
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  call delete(this%qcd)
  call delete(this%hmc_status)
  call delete(this%md_status)

#ifndef _singlePU
  call comlib_finalize
#endif

  return
end subroutine

subroutine print_hmc(this)
!
! Print HMC parameters
!
! this should be called after reading HMC parameters by read method.
!
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  call tic(this%timing_io)

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'("**** Input file ****")')
    write(*,'(8X,"         JOB Input file :",A)')TRIM(this%file_name)
    write(*,*)
    write(*,'("**** Configuration files ****")')
    write(*,'(8X,"Configuration directory :",A)')TRIM(this%config_path)
    write(*,'(8X,"Configuration file name :",A)')TRIM(this%config_file_name_header)
  endif

  call print(this%qcd)

  if (nodeid == 0) then
    write(*,*)
    write(*,'("**** HMC Statistics ****")')
  endif
  call print(this%hmc_status)

  if (nodeid == 0) then
    write(*,*)
    write(*,'("**** MD Statistics ****")')
  endif
  call print(this%md_status)

  if (nodeid == 0) then
    write(*,*)
    write(*,'(80("-"))')
  endif

  call toc(this%timing_io)

  return
end subroutine

subroutine print_stat_hmc(this)
!
! Print HMC statistics at the end of job
!
  use comlib
  implicit none
  type(hmc_algorithm), intent(inout) :: this
  real(DP) :: hmonte_time
  real(DP) :: metropolis_time
  real(DP) :: io_time
  real(DP) :: program_time

  hmonte_time     = get_elapse(this%timing_hmonte)
  metropolis_time = get_elapse(this%timing_metropolis)
  io_time         = get_elapse(this%timing_io)

  call toc(this%timing_program)

  program_time    = get_elapse(this%timing_program)

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'("  Statistics par 1PU")')
    write(*,'("                 Total time (sec) :",F12.4)') program_time
    write(*,'("                 monte time (sec) :",F12.4," (",F8.4," %)")')  &
 &                        hmonte_time,       100*hmonte_time/program_time
    write(*,'("            metropolis time (sec) :",F12.4," (",F8.4," %)")')  &
 &                    metropolis_time,   100*metropolis_time/program_time
    write(*,'("               IO conf time (sec) :",F12.4," (",F8.4," %)")')  &
 &                            io_time,           100*io_time/program_time
  endif

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'("  HMC Statistics")')
  endif

  call print_statistics(this%hmc_status)

  call print_statistics(this%qcd,program_time)

  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif

  return
end subroutine

subroutine read_hmc(this)
!
! Read HMC job parameter file
!
! This should be called after calling new method.
!
  use comlib
  use file_tools_class
  use print_status_class
  implicit none
  type(hmc_algorithm), intent(inout) :: this
  character(len=CHARLEN) :: char
  integer :: iout

  call tic(this%timing_io)
  iout = search_free_file_unit()

  if (nodeid==0) then
    open(iout,file=this%file_name,status='unknown',form='formatted')
    read(iout,'(a)')this%input_file_comment(1)
  endif

!============================  
! read QCD part
!============================  
  call read(this%qcd,iout)

!============================  
! read JOB part
!============================  
  if (nodeid==0) then
    read(iout,'(a)')this%input_file_comment(2)
    read(iout,*)    this%random_number_seed
    read(iout,'(a)')this%config_path
    read(iout,'(a)')this%config_file_name_header
    read(iout,'(a)')this%input_file_comment(3)
  endif

#ifndef _singlePU
  call comlib_bcast(this%random_number_seed,0)
  call comlib_bcast(this%config_path,0)
  call comlib_bcast(this%config_file_name_header,0)
#endif

!============================  
! read HMC part
!============================  
  call read(this%hmc_status,iout)

  if (nodeid==0) then
    read(iout,'(a)')this%input_file_comment(4)
  endif

!============================  
! read MD part
!============================  
  call read(this%md_status,iout)

  if (nodeid==0) then
    close(iout)
  endif

  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif
  write(char,'(1X,A," : read OK ",I3)')TRIM(this%file_name),nodeid
  call print_status(char)

  if (nodeid == 0) write(*,'(80("-"))')

  call toc(this%timing_io)

  !==================================
  ! initialize/load configuration 
  ! for continuation
  ! random number generator 
  ! is initialized here
  !==================================
  call initialize_config_hmc(this)

  return
end subroutine

subroutine save_hmc(this)
!
! Save HMC job parameter
!
! This should be called after the method run ended.
!
  use file_tools_class
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  integer :: iout


  !=======================================
  ! save configuration for continuation
  !=======================================
  call save_contg_hmc(this)

  call tic(this%timing_io)
  iout = search_free_file_unit()

  if (nodeid == 0) then
    open(iout,file=this%file_name,status='unknown',form='formatted')
    write(iout,'(a)')TRIM(this%input_file_comment(1))
  endif

  call save(this%qcd,iout)

  if (nodeid == 0) then
    write(iout,'(a)')TRIM(this%input_file_comment(2))
    write(iout,*)         this%random_number_seed
    write(iout,'(a)')TRIM(this%config_path)
    write(iout,'(a)')TRIM(this%config_file_name_header)
    write(iout,'(a)')TRIM(this%input_file_comment(3))
  endif

  call save(this%hmc_status,iout)

  if (nodeid == 0) then
    write(iout,'(a)')TRIM(this%input_file_comment(4))
  endif

  call save(this%md_status,iout)

  if (nodeid == 0) then
    close(iout)
  endif

  call toc(this%timing_io)

  return
end subroutine

subroutine initialize_config_hmc(this)
!
! Initialize configuration or load cont configuration and make HMC algorithm ready to run.
!
! Random number generator is initialized/loaded here.
!
  use print_status_class
  use random_class
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  character(CHARLEN) :: fname,char
  integer :: irun

  call tic(this%timing_io)

  irun = get_run_number(this%hmc_status)
  fname=fname_cont(irun, &
 &                 this%config_path,                 & 
 &                 this%config_file_name_header)

  select case (irun)
  case (0)   ! cold start with new random number

    call set_trajectory_number(this%hmc_status,0)

    if (nodeid == 0) write(*,'(80("="))')
    if (nodeid == 0) &
    write(*,'("  Cold Start Config")')

    !============================================
    ! initialize config. (cold start)
    !============================================
    call initialize_config(this%qcd,0,fname)

    !===============================
    ! reset and initialize 
    ! random number gernerator
    !===============================
#ifdef _PARALLELCHECK
    call new(g_rand,0,this%random_number_seed,NPU,0)
#else
    call new(g_rand,0,this%random_number_seed,NPU,nodeid)
#endif

  case (:-1) ! continue with new random number

    !===============================
    ! initialize 
    ! random number gernerator
    !===============================
    call new(g_rand,1,this%random_number_seed,NPU,nodeid)

    if (nodeid == 0) write(*,'(80("="))')
    write(char,'("Read Cont. Config from :",a)')TRIM(fname)
    call print_status(char)

    !===============================
    ! total run count is reversed
    !===============================
    call set_run_number(this%hmc_status,-irun)

    !============================================
    ! read cont config.
    !============================================
    call initialize_config(this%qcd,1,fname)

    !===============================
    ! reset total run/traj count
    !===============================
    call        set_run_number(this%hmc_status,0)
    call set_trajectory_number(this%hmc_status,0)

    !===============================
    ! reset and initialize 
    ! random number gernerator
    !===============================
#ifdef _PARALLELCHECK
    call new(g_rand,0,this%random_number_seed,NPU,0)
#else
    call new(g_rand,0,this%random_number_seed,NPU,nodeid)
#endif

  case (1:)  ! continue

    !===============================
    ! initialize 
    ! random number gernerator
    !===============================
    call new(g_rand,1,this%random_number_seed,NPU,nodeid)

    if (nodeid == 0) write(*,'(80("="))')
    write(char,'("Read Cont. Config from :",a)')TRIM(fname)
    call print_status(char)

    !============================================
    ! read cont config.
    ! random number gernerator is resotred.
    !============================================
    call initialize_config(this%qcd,1,fname)

  end select

  call toc(this%timing_io)

  return
end subroutine

subroutine save_config_hmc(this)
!
! Save configuration for measurment
!
! This is called in run mehtod and is private method.
!
  use iso_c_functions, only : f_mkdir
  use file_tools_class
  use hmc_identity
  use print_status_class
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  character(CHARLEN) :: fname
  character(CHARLEN) :: comment,cdate,ctime,czone
  character(CHARLEN) :: cdtnmd
  character(CHARLEN) :: char
  integer :: iout,iflg,itraj
  logical :: flag

  flag = is_save_config(this%hmc_status)

  if (.not. flag) return

  call tic(this%timing_io)

  itraj = get_trajectory_number(this%hmc_status)
  fname=save_path(itraj,this%config_path)

  if (nodeid == 0) then
    write(*,*)
    write(*,'("    make dir :",a)')TRIM(fname)
    call f_mkdir(fname,iflg)
  endif

  fname=fname_save(itraj, &
 &                 this%config_path,       &
 &                 this%config_file_name_header)
  write(char,'("   save conf :",a)')TRIM(fname)
  call print_status(char)

!*********************************************
!* fill character variables with spaces.
!*********************************************
  comment=REPEAT(' ',LEN(comment))
  cdate  =REPEAT(' ',LEN(cdate))
  ctime  =REPEAT(' ',LEN(ctime))
  czone  =REPEAT(' ',LEN(czone))

!*********************************************
!* create comment and created date/time
!*********************************************
  cdtnmd  = TRIM(ADJUSTL(get_parameter_string(this%md_status)))
  comment = TRIM(hmc_vers)
  comment = TRIM(ADJUSTL(comment))//" : "//TRIM(ADJUSTL(cdtnmd))
  call DATE_AND_TIME(DATE=cdate,TIME=ctime,ZONE=czone)
  comment = TRIM(ADJUSTL(comment))
  cdate = TRIM(ADJUSTL(cdate))
  ctime = TRIM(ADJUSTL(ctime))
  czone = TRIM(ADJUSTL(czone))

!**************************************
!* write configuration
!**************************************

  iout = search_free_file_unit()
  open(iout,file=TRIM(ADJUSTL(fname)),status='unknown',form='unformatted')

  call save_config(this%qcd,iout)

!**************************************
!* configuration information
!**************************************
  write(iout) cdate
  write(iout) ctime
  write(iout) czone
  write(iout) comment
  write(iout) this%random_number_seed
  write(iout) itraj
  close(iout)

  if (nodeid == 0) then
    write(*,*)
  endif

  call toc(this%timing_io)

  return
end subroutine

subroutine save_contg_hmc(this)
!
! Save configuration for continuation
!
  use comlib
  use file_tools_class
  use print_status_class
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  character(CHARLEN) :: fname,char
  integer :: iout,irun

  call tic(this%timing_io)

  irun = get_run_number(this%hmc_status)
  iout = search_free_file_unit()
  fname=fname_cont(irun, &
 &                 this%config_path,        &
 &                 this%config_file_name_header)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("                JOB ID :",a)')TRIM(this%job_id)
  endif
  write(char,'(" Write Cont. Config to :",a)')TRIM(fname)
  call print_status(char)

  open(iout,file=fname,status='unknown',form='unformatted',action='write',access='sequential')
  call save_cont_config(this%qcd,iout)
  write(iout)this%job_id
  close(iout)

  if (nodeid == 0) then
    write(*,*)
  endif

  call toc(this%timing_io)

  return
end subroutine

subroutine run_hmc(this)
!
! Run Hybrid Monte Carlo (HMC) Algorithm
!
  implicit none
  type(hmc_algorithm), intent(inout) :: this
  type(timer) :: etime
  character(len=CHARLEN) :: str
  integer :: ihit,isweep
  logical :: flag

  isweep = 0
  do
    !===================================
    ! check sweep/iteration count
    !===================================
    flag = is_loop_end(this%hmc_status)
    if (flag) exit

    !============================================
    ! Hybrid Monte Carlo
    !============================================
    call tic(this%timing_hmonte)
    ihit=0

    !================================
    ! P is generated and ths initial
    ! Hamiltonian is calcd.
    !================================
    call new(etime);call tic(etime)
    call hamil(this%qcd,0)
    call toc(etime)
    call print_log_timing(this%hmc_status,"hamil0:",etime)
#ifdef _REVCHECK
    call rev_check(this%qcd,0)
#endif

    !================================
    ! Molecular Dynamics
    !================================
    call new(etime);call tic(etime)
    call molecular_dynamics(this)
    call toc(etime)
    call print_log_timing(this%hmc_status,"md:",etime)

    !================================
    ! Ths Final Hamiltonian is calcd.
    !================================
    call new(etime);call tic(etime)
    call hamil(this%qcd,1)
    call toc(etime)
    call print_log_timing(this%hmc_status,"hamil1:",etime)

#ifdef _REVCHECK
    call rev_check(this%qcd,1)

    call set_time_step_backward(this%md_status)
    call molecular_dynamics(this)
    call hamil(this%qcd,2)
    call set_time_step_forward(this%md_status)

    call rev_check(this%qcd,2)
#endif
    call toc(this%timing_hmonte)

    !================================
    ! Do Metropolis test;
    ! Q are updated by accept/reject
    ! and P is dropped
    !================================
    call tic(this%timing_metropolis)
    call new(etime);call tic(etime)
    call metropolis(this%qcd,ihit)
    call toc(etime)
    call print_log_timing(this%hmc_status,"metropolis:",etime)
    call toc(this%timing_metropolis)

    !================================
    ! Do measuremet on the fly
    !================================
    call tic(this%timing_io)
    call new(etime);call tic(etime)
    call measurement(this%qcd)
    call toc(etime)
    call print_log_timing(this%hmc_status,"measurement:",etime)
    call toc(this%timing_io)

    !================================
    ! Save configuration for
    ! measurement
    !================================
    call save_config_hmc(this)

    !=======================
    ! increment sweep count
    !=======================
    isweep = isweep + 1
  enddo

  return
end subroutine

subroutine molecular_dynamics(this)
!
! Molecular dynamics integrator for HMC algorithm
!
! this is private method and is used in run method
! 
  implicit none
  type(hmc_algorithm), intent(inout) :: this

  real(DP) :: dt
  integer :: i,idepth

  do
    if (is_loop_end(this%md_status)) exit

        dt =    get_dt(this%md_status)
    idepth = get_depth(this%md_status)

    call set_current_md_depth(this%qcd,idepth)

    select case(is_p(this%md_status))
    case(.TRUE.)
      call update_p(this%qcd,dt)
    case(.FALSE.)
      call update_q(this%qcd,dt)
    end select

  enddo

  return
end subroutine

end module
